/*
 * TpHostGpu.cu
 *
 *  Created on: Apr 2, 2025
 *      Author: elphel
 */
#include <stdexcept>
#include <helper_cuda.h>  // for checkCudaErrors
#include <cuda_runtime.h> // cudaFree
#include <helper_functions.h> // timer functions

//#include "TpParams.h"   // TpHostGpu.h has it
#include "tp_paths.h"
#include "tp_files.h"
#include "tp_utils.h"       // for copyalloc_kernel_gpu
#include "TpHostGpu.h"


TpHostGpu::~TpHostGpu(){
	 hfree(m_host_kern_buf);
	 hfree(m_ftask_data);
	 hfree(m_ftask_data1);

	 hfree(m_correction_vector);
	 hfree(m_rByRDist);
	 gfree(m_gpu_geometry_correction);
	 gfree(m_gpu_correction_vector);
	 gfree(m_gpu_rByRDist);
	 gfree(m_gpu_rot_deriv);

	 for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
		 // Free used host memory, initialized as arrays of pointers
		 hfree(m_gpu_kernels_h[ncam]);
		 hfree(m_gpu_kernel_offsets_h[ncam]);
		 hfree(m_gpu_images_h[ncam]);
		 hfree(m_gpu_clt_h[ncam]);
		 hfree(m_gpu_corr_images_h[ncam]);
		 // Free used GPU memory initialized as **
		 if (m_gpu_kernels)          gfree(m_gpu_kernels[ncam]);
		 if (m_gpu_kernel_offsets)   gfree(m_gpu_kernel_offsets[ncam]);
		 if (m_gpu_images)           gfree(m_gpu_images[ncam]);
		 if (m_gpu_clt)              gfree(m_gpu_clt[ncam]);
		 if (m_gpu_corr_images)      gfree(m_gpu_corr_images[ncam]);
	 }
	 gfree(m_gpu_corrs);
	 gfree(m_gpu_corrs_td);
	 gfree(m_gpu_corrs_combo);
	 gfree(m_gpu_corrs_combo_td);
	 gfree(m_gpu_corr_indices);
	 gfree(m_gpu_corrs_combo_indices);
	 gfree(m_gpu_ftasks);
	 gfree(m_gpu_active_tiles);
	 gfree(m_gpu_num_active);
	 gfree(m_gpu_num_corr_tiles);
	 gfree(m_gpu_texture_indices);
	 gfree(m_gpu_woi);
	 gfree(m_gpu_twh);
	 gfree(m_gpu_num_texture_tiles);
	 gfree(m_gpu_port_offsets);
	 gfree(m_gpu_color_weights);
	 gfree(m_gpu_generate_RBGA_params);
	 gfree(m_gpu_textures);
	 gfree(m_gpu_diff_rgb_combo);
	 gfree(m_gpu_textures_rbga);

	 return;
};
void TpHostGpu::setImageKernels(){
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
       readFloatsFromFile(
       		m_host_kern_buf, // float * data, // allocated array
			m_tpPaths.kernel_file[ncam]); // 			   char *  path) // file path
       m_gpu_kernels_h[ncam] = copyalloc_kernel_gpu(m_host_kern_buf, m_tpParams.kern_size);
       readFloatsFromFile(
    		   m_host_kern_buf, // float * data, // allocated array
			   m_tpPaths.kernel_offs_file[ncam]); // 			   char *  path) // file path
       m_gpu_kernel_offsets_h[ncam] = (struct CltExtra *) copyalloc_kernel_gpu(
    		   m_host_kern_buf,
			   m_tpParams.kern_tiles * (sizeof( struct CltExtra)/sizeof(float)));
    }
    m_gpu_kernels =         copyalloc_pointers_gpu (m_gpu_kernels_h,     m_tpParams.num_cams); // NUM_CAMS);
    m_gpu_kernel_offsets = (struct CltExtra **) copyalloc_pointers_gpu ((float **) m_gpu_kernel_offsets_h, m_tpParams.num_cams); // NUM_CAMS);
}
void TpHostGpu::setCltBuffers(){
	const int slice_size {m_tpParams.tilesy * m_tpParams.tilesx * m_tpParams.num_colors * 4 * m_tpParams.dtt_size * m_tpParams.dtt_size};
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
        // will get results back
    	//gpu_clt_h[ncam] = alloc_kernel_gpu(TILESY * TILESX * tpParams.num_colors * 4 * DTT_SIZE * DTT_SIZE);
        m_gpu_clt_h[ncam] = alloc_kernel_gpu(slice_size);
        printf("Allocating GPU memory, 0x%x floats\n", slice_size) ;
    }
    m_gpu_clt =            copyalloc_pointers_gpu (m_gpu_clt_h,         m_tpParams.num_cams); // NUM_CAMS);
}

void TpHostGpu::setCorrImgBuffers(){
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
        // allocate result images (3x height to accommodate 3 colors
        // Image is extended by 4 pixels each side to avoid checking (mclt tiles extend by 4)
        //host array of pointers to GPU arrays
        m_gpu_corr_images_h[ncam] = alloc_image_gpu(
        		&dstride_rslt,                // size_t* dstride, // in bytes!!
				m_tpParams.img_width + m_tpParams.dtt_size,         // int width,
				m_tpParams.num_colors*(m_tpParams.img_height + m_tpParams.dtt_size));   // int height);
    }
    m_gpu_corr_images =    copyalloc_pointers_gpu (m_gpu_corr_images_h, m_tpParams.num_cams); // NUM_CAMS);
}

void TpHostGpu::setImgBuffers(){
    for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
        readFloatsFromFile(
        		m_host_kern_buf, // float * data, // allocated array
				m_tpPaths.image_files[ncam]); // 			   char *  path) // file path
        m_gpu_images_h[ncam] =  copyalloc_image_gpu(
        		m_host_kern_buf, // float * image_host,
				&dstride,      // size_t* dstride,
				m_tpParams.img_width, // IMG_WIDTH,     // int width,
				m_tpParams.img_height); // IMG_HEIGHT);   // int height);
    }
    m_gpu_images =         copyalloc_pointers_gpu (m_gpu_images_h,      m_tpParams.num_cams); // NUM_CAMS);
}

void TpHostGpu::setGeometryCorrectionBuffers() {
    readFloatsFromFile(
    		(float *) &m_fgeometry_correction, // float * data, // allocated array, no need to free
			m_tpPaths.geometry_correction_file); // char *  path) // file path
    m_rByRDist = readAllFloatsFromFile(
    		m_tpPaths.rByRDist_file, // const char *  path,
    		&m_rByRDist_length);       // int * len_in_floats)
    m_correction_vector =  readAllFloatsFromFile(
    		m_tpPaths.correction_vector_file, // const char *  path,
    		&m_correction_vector_length); // int * len_in_floats)
    m_gpu_geometry_correction =  (struct gc *) copyalloc_kernel_gpu(
    		(float *) &m_fgeometry_correction,
    		sizeof(m_fgeometry_correction)/sizeof(float));
    m_gpu_correction_vector =  (struct corr_vector * ) copyalloc_kernel_gpu(
    		m_correction_vector,
			m_correction_vector_length);
    m_gpu_rByRDist =  copyalloc_kernel_gpu(
    		m_rByRDist,
			m_rByRDist_length);
    checkCudaErrors(cudaMalloc((void **)&m_gpu_rot_deriv, sizeof(trot_deriv)));
}
void TpHostGpu::setCorrelationBuffers(){
    // allocates one correlation kernel per line (15x15 floats), number of rows - number of tiles * number of pairs
    m_gpu_corrs = alloc_image_gpu(
    		&dstride_corr,                  // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			m_tpParams.corr_length,                      // int width,
			m_tpParams.num_pairs * m_tpParams.tilesx * m_tpParams.tilesy);   // int height);
    // read channel images (assuming host_kern_buf size > image size, reusing it)
    // allocate all other correlation data, some may be
    m_gpu_corrs_td = alloc_image_gpu(
    		&dstride_corr_td,                                                 // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			4 * m_tpParams.dtt_size *m_tpParams.dtt_size,                     // int width,
			m_tpParams.num_pairs * m_tpParams.tilesx * m_tpParams.tilesy);    // int height);

    m_gpu_corrs_combo = alloc_image_gpu(
    		&dstride_corr_combo,                                              // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			m_tpParams.corr_length,                                             // int width,
			m_tpParams.tilesx * m_tpParams.tilesy);                           // int height);

    m_gpu_corrs_combo_td = alloc_image_gpu(
    		&dstride_corr_combo_td,                                           // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
			4 * m_tpParams.dtt_size *m_tpParams.dtt_size,                     // int width,
			m_tpParams.tilesx * m_tpParams.tilesy);                           // int height);
    // just allocate
   checkCudaErrors (cudaMalloc((void **)&m_gpu_corr_indices,  m_tpParams.num_pairs * m_tpParams.tilesx * m_tpParams.tilesy * sizeof(int)));
   checkCudaErrors (cudaMalloc((void **)&m_gpu_corrs_combo_indices,            m_tpParams.tilesx * m_tpParams.tilesy * sizeof(int)));
}

void TpHostGpu::setTasks(const float target_disparity, const float scale){
	// allocate m_ftask_data
	m_ftask_data  =   (float *) malloc( m_tpParams.tilesx *  m_tpParams.tilesy * m_tpParams.task_size * sizeof(float));
//    float * ftask_data1  =  (float *) malloc(TILESX * TILESY * tpParams.task_size * sizeof(float));

	   // tasks for all tiles
	    for (int ty = 0; ty < m_tpParams.tilesy ; ty++){
	        for (int tx = 0; tx < m_tpParams.tilesx ; tx++){
	            int nt = ty * m_tpParams.tilesx  + tx;
	            int task_task = (1 << m_tpParams.task_inter_en) | (1 << m_tpParams.task_corr_en) | (1 << m_tpParams.task_text_en); // just 1 bit, correlation selection is defined by common corr_sel bits
	            int task_txy = tx + (ty << 16);
	//            float task_target_disparity = DBG_DISPARITY; // disparity for which to calculate offsets (not needed in Java)
	            float * tp = m_ftask_data + m_tpParams.task_size * nt;
	            *(tp + m_tpParams.tp_task_task_offset) =    *(float *) &task_task;
	            *(tp + m_tpParams.tp_task_txy_offset) =      *(float *) &task_txy;
	            *(tp + m_tpParams.tp_task_disparity_offset) =  target_disparity;
	//            tp += 2; // skip centerX, centerY
	            *(tp + m_tpParams.tp_task_scale_offset) = scale; //  0; // 0.5f; // ,0; // scale, 0 - old way, just set
	            tp+= m_tpParams.tp_task_xy_offset;
	            for (int ncam = 0; ncam < m_tpParams.num_cams; ncam++) {
	            	*(tp++) = m_tile_coords_h[ncam][nt][0];
	            	*(tp++) = m_tile_coords_h[ncam][nt][1];
	            }
	        }
	    }
	    m_gpu_ftasks = (float  *) copyalloc_kernel_gpu(m_ftask_data, m_tpParams.tp_tasks_size * m_tpParams.task_size); // (sizeof(struct tp_task)/sizeof(float)));
}
void TpHostGpu::setTextures(){
	if (!m_ftask_data) { // ftask_data is not initialized
		throw std::runtime_error("Error: m_ftask_data is not initialized ");
	}
	m_tile_texture_layers = (m_tpParams.texture_colors + 1 + (m_tpParams.keep_texture_weights? (m_tpParams.num_cams + m_tpParams.texture_colors + 1): 0));
	m_tile_texture_size = m_tile_texture_layers *256;
	 // task data should be initialized
	m_num_textures=0;
    for (int ty = 0; ty < m_tpParams.tilesy; ty++){
    	for (int tx = 0; tx < m_tpParams.tilesx; tx++){
    		int nt = ty * m_tpParams.tilesx + tx;
            float *tp = m_ftask_data + m_tpParams.task_size * nt;
    		int cm = (*(int *) tp) & m_tpParams.task_texture_bits; // non-zero any of 4 lower task bits
    		if (cm){
    			m_texture_indices[m_num_textures++] = (nt << m_tpParams.text_ntile_shift) | (1 << m_tpParams.list_texture_bit); // setting  0x80 in texture indices
    		}
    	}
    }
    // num_textures now has the total number of textures
    // copy corr_indices to gpu
    m_gpu_texture_indices = (int  *) copyalloc_kernel_gpu(
    		(float * ) m_texture_indices,
			m_num_textures,
			m_tpParams.tilesx * m_tpParams.tilesya); // number of rows - multiple of 4
    m_gpu_textures = alloc_image_gpu(
    		&dstride_textures,                      // in bytes ! for one rgba/ya 16x16 tile
			m_tile_texture_size,                    // int width (floats),
			m_tpParams.tilesx * m_tpParams.tilesy); // int height);
}

void TpHostGpu::setRGBA(){
	/*
	// appears in setTextures
    m_tile_texture_layers = (m_tpParams.texture_colors + 1 + (m_tpParams.keep_texture_weights? (m_tpParams.num_cams + m_tpParams.texture_colors + 1): 0));
    m_tile_texture_size = m_tile_texture_layers *256;
    */
    m_rgba_width =   (m_tpParams.tilesx + 1) * m_tpParams.dtt_size;
    m_rgba_height =  (m_tpParams.tilesy + 1) * m_tpParams.dtt_size;;
    m_rbga_slices =  m_tpParams.texture_colors + 1; // 4/1
    if (m_tpParams.keep_texture_weights & 2){
    	m_rbga_slices += m_tpParams.texture_colors * m_tpParams.num_cams;
    }
    checkCudaErrors(cudaMalloc((void **)&m_gpu_woi,               4 * sizeof(float))); // RGBA
    checkCudaErrors(cudaMalloc((void **)&m_gpu_twh,               2 * sizeof(float))); // RGBA, dynamic
    checkCudaErrors(cudaMalloc((void **)&m_gpu_num_texture_tiles, 8 * sizeof(float))); // for each subsequence - number of non-border, RGBA all
	m_gpu_port_offsets =         (float *) copyalloc_kernel_gpu((float * ) m_tpParams.port_offsets, m_tpParams.num_cams * 2); // num_ports * 2);
    m_gpu_color_weights =        (float *) copyalloc_kernel_gpu((float * ) m_tpParams.color_weights, sizeof(m_tpParams.color_weights));
    m_gpu_generate_RBGA_params = (float *) copyalloc_kernel_gpu((float * ) m_tpParams.generate_RBGA_params, sizeof(m_tpParams.generate_RBGA_params));

    m_gpu_textures_rbga = alloc_image_gpu(
   		&dstride_textures_rbga,              // in bytes ! for one rgba/ya 16x16 tile
			m_rgba_width,              // int width (floats),
			m_rgba_height * m_rbga_slices);               // int height);
    checkCudaErrors(cudaMalloc((void **)&m_gpu_diff_rgb_combo,  m_tpParams.tilesx * m_tpParams.tilesy * m_tpParams.num_cams *
    		(m_tpParams.num_colors + 1) * sizeof(float)));
}

void TpHostGpu::testCorrelate2DIntra(int num_runs){
	int numIterations = m_tpParams.debug_tile ? 1 : num_runs;
	int i0 =            m_tpParams.debug_tile ? 0 : -1;
	if (!m_gpu_clt){
		throw std::runtime_error("Error: m_gpu_clt is not initialized ");
	}
	if (!m_gpu_ftasks) {
		throw std::runtime_error("Error: m_gpu_ftasks is not initialized ");
	}
	if (!m_gpu_corrs) {
		throw std::runtime_error("Error: m_gpu_corrs is not initialized ");
	}
    StopWatchInterface *timerCORR = 0;
    sdkCreateTimer(&timerCORR);
    for (int i = i0; i < numIterations; i++)
    {
    	if (i == 0)
    	{
    		checkCudaErrors(cudaDeviceSynchronize());
    		sdkResetTimer(&timerCORR);
    		sdkStartTimer(&timerCORR);
    	}
    	correlate2D<<<1,1>>>(
    			m_tpParams.num_cams,         // int               num_cams,
				m_tpParams.sel_pairs[0],     // int               sel_pairs0           // unused bits should be 0
				m_tpParams.sel_pairs[1],     // int               sel_pairs1,           // unused bits should be 0
				m_tpParams.sel_pairs[2],     // int               sel_pairs2,           // unused bits should be 0
				m_tpParams.sel_pairs[3],     // int               sel_pairs3,           // unused bits should be 0
				m_gpu_clt,                   // float          ** gpu_clt,            // [NUM_CAMS] ->[TILESY][TILESX][NUM_COLORS][DTT_SIZE*DTT_SIZE]
				m_tpParams.num_colors,       // int               colors,             // number of colors (3/1)
				m_tpParams.color_weights[0], // 0.25,  // float             scale0,             // scale for R
				m_tpParams.color_weights[1], // 0.25,  // float             scale1,             // scale for B
				m_tpParams.color_weights[2], // 0.5,   // float             scale2,             // scale for G
				m_tpParams.fat_zero * m_tpParams.fat_zero,                // float             fat_zero2,           // here - absolute
				m_gpu_ftasks,                // float            * gpu_ftasks,         // flattened tasks, 27 floats for quad EO, 99 floats for LWIR16
				m_tpParams.tp_tasks_size,    // int               num_tiles) // number of tiles in task
				m_tpParams.tilesx,           // int               tilesx,             // number of tile rows
				m_gpu_corr_indices,          // int             * gpu_corr_indices,   // packed tile+pair
				m_gpu_num_corr_tiles,        // int             * pnum_corr_tiles,    // pointer to a number of correlation tiles to process
				dstride_corr/sizeof(float),  // const size_t      corr_stride,        // in floats
				m_tpParams.corr_out_rad,     // int               corr_radius,        // radius of the output correlation (7 for 15x15)
				m_gpu_corrs);                // float           * gpu_corrs);          // correlation output data

    	getLastCudaError("Kernel failure");
    	checkCudaErrors(cudaDeviceSynchronize());
    	printf("test pass: %d\n",i);
    }
    float avgTimeCORR = (float)sdkGetTimerValue(&timerCORR) / (float)numIterations;
    sdkStopTimer(&timerCORR);
    sdkDeleteTimer(&timerCORR);
    int num_corrs{}; // will get data from the gpu memory
    checkCudaErrors(cudaMemcpy(
    		&num_corrs,
			m_gpu_num_corr_tiles,
			sizeof(int),
			cudaMemcpyDeviceToHost));
    printf("Average CORR run time =%f ms, num cor tiles (new) = %d\n",  avgTimeCORR, num_corrs);
    saveIntraCorrFile(
    		m_tpPaths.result_corr_file, // const char * path,
			"phase correlation data",   // const char * prompt,
			num_corrs, // int          num_corrs,
			m_gpu_corrs, // float *      gpu_corrs,
			m_gpu_corr_indices, // int *        gpu_corr_indices)
			16); // //int          num_sel_sensors) { // only for interscene

}


void TpHostGpu::saveIntraCorrFile(
		const char * path,
		const char * prompt,
		int          num_corrs,
		float *      gpu_corrs,
		int *        gpu_corr_indices,
		int          num_sel_sensors) { // only for interscene
	if (!path)	return;
	int rslt_corr_length =   num_corrs * m_tpParams.corr_length;
	int corr_img_size =    m_tpParams.num_corr_indices * 16 * 16; // NAN
//	float * corr_img = (float *)malloc(corr_img_size * sizeof(float));
	float * cpu_corr = (float *)malloc(rslt_corr_length * sizeof(float));
	int *   cpu_corr_indices = (int *) malloc(m_tpParams.num_corr_indices * sizeof(int));
	checkCudaErrors(cudaMemcpy2D(
			cpu_corr,
			m_tpParams.corr_length * sizeof(float),
			gpu_corrs,
			dstride_corr,
			m_tpParams.corr_length * sizeof(float),
			num_corrs,
			cudaMemcpyDeviceToHost));
	checkCudaErrors(cudaMemcpy(
			cpu_corr_indices,
			gpu_corr_indices,
			m_tpParams.num_corr_indices * sizeof(int),
			cudaMemcpyDeviceToHost));
	float * corr_img = 	getCorrImg(
			corr_img_size, // int     corr_img_size,
			cpu_corr_indices, // int *   cpu_corr_indices,
			cpu_corr, // float * cpu_corr,
			num_sel_sensors); // int     num_sel_sensors)
    printf("Writing %s to %s, width = %d, height=%d, slices=%d, length=%ld bytes\n",
    		prompt, path, (m_tpParams.tilesx * 16),(m_tpParams.tilesya * 16), m_tpParams.num_pairs, (corr_img_size * sizeof(float)) ) ;
    writeFloatsToFile(
    		corr_img,       // float *       data, // allocated array
			corr_img_size,  // int           size, // length in elements
			path);          // 			   const char *  path) // file path
    free (cpu_corr);
    free (cpu_corr_indices);
    free (corr_img);
}

/*

    printf("Writing interscene phase correlation data to %s, width = %d, height=%d, slices=%d, length=%ld bytes\n",
    		tpPaths.result_inter_td_norm_file, (TILESX*16),(TILESYA*16), num_pairs_inter, (corr_img_size * sizeof(float)) ) ;
    writeFloatsToFile(
    		corr_img,                  // float *       data, // allocated array
			corr_img_size,             // int           size, // length in elements
			tpPaths.result_inter_td_norm_file); // 			   const char *  path) // file path
 */

float * TpHostGpu::getCorrImg(
		int     corr_img_size,
		int *   cpu_corr_indices,
		float * cpu_corr,
		int     num_sel_sensors){
	float * corr_img = (float *)malloc(corr_img_size * sizeof(float));
	for (int i = 0; i < corr_img_size; i++){
		corr_img[i] = NAN;
	}
    for (int ict = 0; ict < m_tpParams.num_corr_indices; ict++){
    	int ctt = ( cpu_corr_indices[ict] >>  m_tpParams.corr_ntile_shift); // CORR_NTILE_SHIFT);
    	int cpair = cpu_corr_indices[ict] & ((1 << m_tpParams.corr_ntile_shift) - 1);
    	if (cpair == 0xff){ // Was not here - only for interscene. Will it hurt?
    		cpair = num_sel_sensors;
    	}
    	int ty = ctt / m_tpParams.tilesx;
    	int tx = ctt % m_tpParams.tilesx;
    	int src_offs0 = ict * m_tpParams.corr_length;
    	int dst_offs0 = cpair * (m_tpParams.num_tiles * 16 * 16) +  (ty * 16 * m_tpParams.tilesx * 16) + (tx * 16);
    	for (int iy = 0; iy < m_tpParams.corr_size; iy++){
    		int src_offs = src_offs0 + iy * m_tpParams.corr_size; // ict * num_pairs * corr_size * corr_size;
    		int dst_offs = dst_offs0 + iy * (m_tpParams.tilesx * 16);
    		for (int ix = 0; ix < m_tpParams.corr_size; ix++){
    			corr_img[dst_offs++] = cpu_corr[src_offs++];
    		}
    	}
    }
	return corr_img;
}

float * TpHostGpu::getCorrTdImg(
		int     corr_img_size,
		int *   cpu_corr_indices,
		float * cpu_corr_td,
		int     num_sel_sensors){
	float * corr_img = (float *)malloc(corr_img_size * sizeof(float));
	for (int i = 0; i < corr_img_size; i++){
		corr_img[i] = NAN;
	}
    for (int ict = 0; ict < m_tpParams.num_corr_indices; ict++){
    	int ctt = ( cpu_corr_indices[ict] >>  m_tpParams.corr_ntile_shift); // CORR_NTILE_SHIFT);
    	int cpair = cpu_corr_indices[ict] & ((1 << m_tpParams.corr_ntile_shift) - 1);
    	if (cpair == 0xff){ // Was not here - only for interscene. Will it hurt?
    		cpair = num_sel_sensors;
    	}
    	int ty = ctt / m_tpParams.tilesx;
    	int tx = ctt % m_tpParams.tilesx;
    	int src_offs0 = ict * m_tpParams.corr_length;
    	int dst_offs0 = cpair * (m_tpParams.num_tiles * 16 * 16) +  (ty * 16 * m_tpParams.tilesx * 16) + (tx * 16);

    	for (int iquad = 0; iquad < 4; iquad ++) {
    		int iqy = (iquad >> 1) & 1;
    		int iqx = (iquad >> 0) & 1;
    		for (int iy = 0; iy < m_tpParams.dtt_size; iy++){
    			int src_offs = src_offs0 + iy * m_tpParams.dtt_size + iquad * m_tpParams.dtt_size * m_tpParams.dtt_size;
    			int dst_offs = dst_offs0 + (iy + m_tpParams.dtt_size * iqy)* (m_tpParams.tilesx * 16) + iqx * m_tpParams.dtt_size;
    			for (int ix = 0; ix < m_tpParams.dtt_size; ix++){
    				corr_img[dst_offs++] = cpu_corr_td[src_offs++];
    			}
    		}
    	}
    }
	return corr_img;
}



void TpHostGpu::saveInterCorrFile(
		const char * path,   // "clt/aux_interscene-TD.raw"
		const char * prompt, // "interscene phase correlation"
		int          num_corrs,
		float      * gpu_corrs_td,
		int        * gpu_corr_indices,
		int          num_sel_sensors){
	if (!path)	return;
	int corr_img_size =    m_tpParams.num_corr_indices * 16 * 16; // NAN
 	int rslt_corr_size_td =   num_corrs *  m_tpParams.dtt_size2 * m_tpParams.dtt_size2;
 	float * cpu_corr_td = (float *)malloc(rslt_corr_size_td * sizeof(float));
 	int     dtile_bytes = (m_tpParams.dtt_size2*m_tpParams.dtt_size2) * sizeof(float);
	int *   cpu_corr_indices = (int *) malloc(m_tpParams.num_corr_indices * sizeof(int));

    checkCudaErrors(cudaMemcpy2D(
    		cpu_corr_td,
			dtile_bytes,
			gpu_corrs_td,
			dstride_corr_td,
			dtile_bytes,
			num_corrs,
			cudaMemcpyDeviceToHost));
    checkCudaErrors(cudaMemcpy(
    		cpu_corr_indices,
			gpu_corr_indices,
			m_tpParams.num_corr_indices * sizeof(int),
			cudaMemcpyDeviceToHost));

	float * corr_img = 	getCorrTdImg(
			corr_img_size,    // int     corr_img_size,
			cpu_corr_indices, // int *   cpu_corr_indices,
			gpu_corrs_td,     // float * cpu_corr,
			num_sel_sensors); // int     num_sel_sensors)
    printf("Writing %s TD data to %s\n", prompt, path);
    writeFloatsToFile(
    		corr_img,        // float *       data, // allocated array
			corr_img_size,   // int           size, // length in elements
			path);           // 			   const char *  path) // file path
    free(cpu_corr_indices);
	free (corr_img);
    free (cpu_corr_td);
}

void TpHostGpu::saveInterCorrIndicesFile(
		const char * path,   //  "clt/aux_inter-indices.raw"
		const char * prompt, //  "interscene indices"
		int * gpu_corr_indices,
		int num_sel_sensors){
	if (!path)	return;
	int *   cpu_corr_indices = (int *) malloc(m_tpParams.num_corr_indices * sizeof(int));
    checkCudaErrors(cudaMemcpy(
    		cpu_corr_indices,
			gpu_corr_indices,
			m_tpParams.num_corr_indices * sizeof(int),
			cudaMemcpyDeviceToHost));
    int corr_index_img_length = m_tpParams.tilesx * m_tpParams.tilesy * (num_sel_sensors+1) ;
    float *corr_index_img =  (float *)malloc(corr_index_img_length * sizeof(float));
    for (int i = 0; i < corr_index_img_length; i++){
    	corr_index_img[i] = NAN;
    }
    for (int ict = 0; ict < m_tpParams.num_corr_indices; ict++){
    	int ctt = ( cpu_corr_indices[ict] >>  m_tpParams.corr_ntile_shift); // CORR_NTILE_SHIFT);
    	int cpair = cpu_corr_indices[ict] & ((1 << m_tpParams.corr_ntile_shift) - 1);
    	if (cpair == 0xff){
    		cpair = num_sel_sensors;
    	}
    	int ty = ctt / m_tpParams.tilesx;
    	int tx = ctt % m_tpParams.tilesx;
    	corr_index_img[cpair * m_tpParams.tilesx * m_tpParams.tilesy + m_tpParams.tilesx * ty + tx] =  ict; // cpu_corr_indices[ict];
    }
    printf("Writing %s to %s\n",prompt,path);
    writeFloatsToFile(
    		corr_index_img,                  // float *       data, // allocated array
			corr_index_img_length,           // int           size, // length in elements
			"clt/aux_inter-indices.raw"); // 			   const char *  path) // file path
    free (corr_index_img);
    free(cpu_corr_indices);
}



void TpHostGpu::hfree(float *              p) {if (p) free(p); p = {};}
void TpHostGpu::hfree(struct CltExtra *    p) {if (p) free(p); p = {};}
void TpHostGpu::gfree(float *              p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(int *                p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct CltExtra *    p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct gc *          p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct corr_vector * p) {if (p) checkCudaErrors(cudaFree(p));p = {};}
void TpHostGpu::gfree(struct trot_deriv  * p) {if (p) checkCudaErrors(cudaFree(p));p = {};}


