/*
 * TpHostGpu.h
 *
 *  Created on: Apr 2, 2025
 *      Author: elphel
 */

#ifndef SRC_TPHOSTGPU_H_
#define SRC_TPHOSTGPU_H_

#include "geometry_correction.h"
#include "TpParams.h"
#include "TileProcessor.h"

class TpHostGpu{
	static constexpr int m_num_cams_lwir =   16;
	static constexpr int m_num_cams_rgb =     4;
	static constexpr int max_num_cams {std::max(m_num_cams_rgb,m_num_cams_lwir)}; // it is always 16 element, RGB uses only first 4
	/*
	 *
	 */

public:
	TpParams& m_tpParams;
	TpPaths& m_tpPaths;
private:
	// no need to free
    float               m_tile_coords_h        [max_num_cams][TpParams::tilesx*TpParams::tilesy][2]; // [TILESX * TILESY][2];
    struct gc           m_fgeometry_correction;
    int                 m_correction_vector_length{};
    int                 m_rByRDist_length{};
    int                 m_texture_indices [TpParams::tilesx * TpParams::tilesya]; //     [TILESX*TILESYA];
    int                 m_num_textures{};
    int                 m_tile_texture_layers{};
    int                 m_tile_texture_size{};

    int                 m_rgba_width{};  //  =   (TILESX+1) * DTT_SIZE;
    int                 m_rgba_height{}; //  =  (TILESY+1) * DTT_SIZE;
    int                 m_rbga_slices{}; //  =  tpParams.texture_colors + 1; // 4/1


	// need to free
	float *             m_host_kern_buf{};
	float *             m_ftask_data{};
	float *             m_ftask_data1{};
    float *             m_gpu_ftasks{};
    int *               m_gpu_active_tiles{};   // tasks
    int *               m_gpu_num_active{};     // tasks
    int *               m_gpu_num_corr_tiles{}; // correlations


	// host memory
	// dstride* - size in byte to be passed to the GPU kernels
	size_t              dstride{};
    size_t              dstride_rslt{};          // in bytes !
    size_t              dstride_corr{};          // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
    // in the future, dstride_corr can reuse that of dstride_corr_td?
    size_t              dstride_corr_td{};       // in bytes ! for one 2d phase correlation (padded 4x8x8x4 bytes)
    size_t              dstride_corr_combo{};    // in bytes ! for one 2d phase correlation (padded 15x15x4 bytes)
    size_t              dstride_corr_combo_td{}; // in bytes ! for one 2d phase correlation (padded 4x8x8x4 bytes)
    size_t              dstride_textures{}; // in bytes ! for one rgba/ya 16x16 tile
    size_t              dstride_textures_rbga{}; // in bytes ! for one rgba/ya 16x16 tile

	float *             m_gpu_kernels_h[max_num_cams]{};
	struct CltExtra *   m_gpu_kernel_offsets_h[max_num_cams]{};
	float *             m_gpu_images_h[max_num_cams]{};
	float *             m_gpu_clt_h[max_num_cams]{};
	float *             m_gpu_corr_images_h[max_num_cams]{};
    float *             m_correction_vector{};
    float *             m_rByRDist{};
    int   *             m_gpu_texture_indices{};



	// GPU memory
	float **            m_gpu_kernels{};
	struct CltExtra  ** m_gpu_kernel_offsets{};
	float **            m_gpu_images{};
	float **            m_gpu_clt{};
	float **            m_gpu_corr_images{};

    struct gc          * m_gpu_geometry_correction{};
    struct corr_vector * m_gpu_correction_vector{};
    float              * m_gpu_rByRDist{};
    struct trot_deriv  * m_gpu_rot_deriv{};

    // correlations device memory
    float              * m_gpu_corrs{};               // correlation tiles (per tile, per pair) in pixel domain
    float              * m_gpu_corrs_td{};            // correlation tiles (per tile, per pair) in transform domain
    float              * m_gpu_corrs_combo{};         // correlation tiles combined (1 per tile), pixel domain
    float              * m_gpu_corrs_combo_td{};      // correlation tiles combined (1 per tile), transform domain
    int                * m_gpu_corr_indices{};        // shared by gpu_corrs gpu_corrs_td
    int                * m_gpu_corrs_combo_indices{}; // shared by gpu_corrs_combo and gpu_corrs_combo_td

    float *             m_gpu_textures{};
    float *             m_gpu_diff_rgb_combo{};
    float *             m_gpu_textures_rbga{};

    int   *             m_gpu_woi{};
    int   *             m_gpu_twh{};
    int   *             m_gpu_num_texture_tiles{};

    float *             m_gpu_port_offsets{};
    float *             m_gpu_color_weights{};
    float *             m_gpu_generate_RBGA_params{};

public:
	TpHostGpu(TpParams& tpParams, TpPaths& tpPaths)
	:m_tpParams{tpParams}
	,m_tpPaths{tpPaths}
	,m_host_kern_buf{(float *) malloc(tpParams.kern_size * sizeof(float))}
	{};
	~TpHostGpu();
	void setImageKernels();
	void setCltBuffers();
	void setCorrImgBuffers();
	void setImgBuffers();
	void setGeometryCorrectionBuffers();
	void setCorrelationBuffers();
	void setTasks(const float target_disparity, const float scale);
	void setTextures();
	void setRGBA();
	void testCorrelate2DIntra(int num_runs);
	// for both intra and inter!
	void saveIntraCorrFile(const char * path, const char * prompt, int num_corrs, float * gpu_corrs, int * gpu_corr_indices, int num_sel_sensors);
	void saveInterCorrFile(const char * path, const char * prompt, int num_corrs, float  * gpu_corrs_td,  int * gpu_corr_indices, int num_sel_sensors);
	void saveInterCorrIndicesFile(const char * path, const char * prompt, int * gpu_corr_indices, int num_sel_sensors);

private:
	float * getCorrImg(int corr_img_size, int * cpu_corr_indices, float * cpu_corr, int num_sel_sensors);
	float * getCorrTdImg(int corr_img_size, int * cpu_corr_indices, float * cpu_corr_td, int num_sel_sensors);
	void hfree(float * p); // {if (p) free (p);}
	void hfree(struct CltExtra * p);
	void gfree(float * p);
	void gfree(int * p);
	void gfree(struct CltExtra * p);

	void gfree(struct gc * p);
	void gfree(struct corr_vector * p);
	void gfree(struct trot_deriv * p);

};



#endif /* SRC_TPHOSTGPU_H_ */
